from b2luigi.core import utils
from b2luigi.core.settings import get_setting

import ROOT
import luigi

import os
import sys
import collections
import enum
import contextlib
import subprocess

from b2luigi.helpers import TaskWithDataSupport


class NonParseableParameter(luigi.Parameter):
    def parse(self, value):
        raise NotImplementedError()


class FunctionParameter(NonParseableParameter):
    def serialize(self, value):
        return value.__name__


class ROOTLocalTarget(luigi.LocalTarget):
    def exists(self):
        if not super().exists():
            return False

        path = self.path
        tfile = ROOT.TFile.Open(path)
        return tfile and len(tfile.GetListOfKeys()) > 0


class Task(luigi.Task, TaskWithDataSupport):
    git_hash = luigi.Parameter(default=utils.get_basf2_git_hash())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        filename = os.path.realpath(sys.argv[0])
        log_folder = os.path.join(os.path.dirname(filename), "logs")
        stdout_file_name = self.get_output_file_name(self.get_task_family() + "_stdout", create_folder=True,
                                                     result_path=log_folder)

        stderr_file_name = self.get_output_file_name(self.get_task_family() + "_stderr", create_folder=True,
                                                     result_path=log_folder)

        self.log_files = {"stdout":  stdout_file_name,
                          "stderr": stderr_file_name,
                          "log_folder": log_folder}

        self.check_complete = True

    def get_filled_params(self):
        return {key: getattr(self, key) for key, _ in self.get_params()}

    def add_to_output(self, output_file_name):
        assert isinstance(output_file_name, str)
        return {output_file_name: self.get_output_file_target(output_file_name, create_folder=True)}

    def get_log_output_files(self):
        return self.log_files

    def get_transposed_input_file_names(self):
        return self.get_input_file_names()

    def get_input_file_names(self):
        return_dict = collections.defaultdict(list)

        for i in self.input():
            file_names = utils.flatten_to_file_paths(utils.flatten_to_dict(i))

            for key, file_name in file_names.items():
                return_dict[key].append(file_name)

        return {key: value for key, value in return_dict.items()}

    def get_output_file_names(self):
        return utils.flatten_to_file_paths(
            utils.flatten_to_dict(self.output())
            )

    def get_output_file_target(self, *args, **kwargs):
        file_name = self.get_output_file_name(*args, **kwargs)
        if os.path.splitext(file_name)[-1] == ".root":
            return ROOTLocalTarget(file_name)
        else:
            return luigi.LocalTarget(file_name)

    def get_serialized_parameters(self, guess_type=False):
        serialized_parameters = collections.OrderedDict()

        for key, parameter in self.get_params():
            if not parameter.significant:
                continue

            value = getattr(self, key)

            if not guess_type or isinstance(value, enum.Enum):
                value = parameter.serialize(value)

            # TODO: this is a bit unfortunate....
            if isinstance(value, str) and "/" in value:
                value = os.path.splitext(os.path.split(value)[-1])[0]

            serialized_parameters[key] = value

        # Git hash should go to the front
        return_dict = collections.OrderedDict()
        return_dict["git_hash"] = serialized_parameters["git_hash"]

        for key, value in serialized_parameters.items():
            return_dict[key] = value

        return return_dict

    def get_output_file_name(self, base_filename, create_folder=False, result_path=None):
        serialized_parameters = self.get_serialized_parameters()

        if not result_path:
            result_path = get_setting("result_path", ".")

        filename = os.path.join(result_path,
                                *[f"{key}={value}" for key,
                                    value in serialized_parameters.items()],
                                base_filename)

        if create_folder:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        return filename

    def complete(self):
        if not super().complete():
            return False

        if not self.check_complete:
            return True

        requires = self.requires()

        try:
            if not requires.complete():
                return False
        except AttributeError:
            for task in requires:
                if not task.complete():
                    return False

        return True

    def create_output_dirs(self):
        output_list = utils.flatten_to_dict(self.output())
        output_list = output_list.values()

        for output in output_list:
            output.makedirs()

class ExternalTask(Task, luigi.ExternalTask):
    pass

class WrapperTask(Task, luigi.WrapperTask):
    pass


class DispatchableTask(Task):
    cmd_prefix = []

    def process(self):
        raise NotImplementedError

    def run_local(self):
        self.create_output_dirs()

        self.process()

    def run_remote(self):
        env_list = self._prepare_env()
        self.dispatch(os.path.realpath(sys.argv[0]), env_list)

    def run(self):
        if os.environ.get("B2LUIGI_EXECUTION", False) or not get_setting("dispatch", True):
            self.run_local()
        else:
            self.run_remote()

    def dispatch(self, filename, env):
        stdout_file_name = self.log_files["stdout"]
        stderr_file_name = self.log_files["stderr"]

        with contextlib.suppress(FileNotFoundError):
            os.remove(stdout_file_name)
        with contextlib.suppress(FileNotFoundError):
            os.remove(stderr_file_name)

        process_env = os.environ.copy()
        process_env.update(env)

        with open(stdout_file_name, "w") as stdout_file:
            with open(stderr_file_name, "w") as stderr_file:
                return_code = subprocess.call(self.cmd_prefix + [sys.executable, os.path.basename(filename)],
                                              stdout=stdout_file, stderr=stderr_file,
                                              env=process_env, cwd=os.path.dirname(filename))

        if return_code:
            raise RuntimeError(
                f"Basf2 execution failed with return code {return_code}")

    def _prepare_env(self):
        env = {
            "B2LUIGI_TASK": utils.encode_value(self.__class__),
            "B2LUIGI_EXECUTION": "1"
        }

        for key, parameter in self.get_params():
            value = getattr(self, key)

            env[f"{utils.PREFIX}{key}"] = utils.encode_value(value)
        return env


def run_task_from_env():
    task_class_name = os.environ["B2LUIGI_TASK"]
    task_class = utils.decode_value(task_class_name)

    params = {}

    for key, value in os.environ.items():
        if not key.startswith(utils.PREFIX):
            continue

        params[key.replace(utils.PREFIX, "")] = utils.decode_value(value)

    task = task_class(**params)
    task.run()





#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['b2luigi', 'b2luigi.cli', 'b2luigi.core']

package_data = \
{'': ['*']}

install_requires = \
['luigi', 'parse', 'GitPython']

setup(name='b2luigi',
      version='0.1.2',
      description='Task scheduling and batch running for basf2 jobs made simple',
      author='Nils Braun',
      author_email='nils.braun@kit.edu',
      url='https://github.com/nils-braun/b2luigi',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )



from collections import namedtuple
from glob import glob
import enum

from b2luigi.core.utils import product_dict, fill_kwargs_with_lists

import b2luigi

from parse import parse


MDST_DIR_STRUCTURE = "/hsm/belle2/bdata/Data/release-{p.release}/DB{p.database:08d}/prod{p.prod:08d}/" + \
                     "e{p.experiment_number:04d}/4S/r{p.run_number:05d}/all/mdst/sub00/" + \
                     "mdst.{p.prefix}.{p.experiment_number:04d}.{p.run_number:05d}.{p.file_name}.root"
CDST_DIR_STRUCTURE = "/hsm/belle2/bdata/Data/release-{p.release}/DB{p.database:08d}/prod{p.prod:08d}/" + \
                     "e{p.experiment_number:04d}/4S/r{p.run_number:05d}/all/cdst/sub00/" + \
                     "cdst.{p.prefix}.{p.experiment_number:04d}.{p.run_number:05d}.{p.file_name}.root"
RAW_DIR_STRUCTURE = "/ghi/fs01/belle2/bdata/Data/Raw/e{p.experiment_number:04d}/r{p.run_number:05d}/sub00/" + \
                    "{p.prefix}.{p.experiment_number:04d}.{p.run_number:05d}.{p.file_name}.root"


class DataMode(enum.Enum):
    raw = "raw"
    mdst = "mdst"
    cdst = "cdst"


class DataTask(b2luigi.ExternalTask):
    data_mode = b2luigi.EnumParameter(enum=DataMode)
    experiment_number = b2luigi.IntParameter()
    run_number = b2luigi.IntParameter()
    prefix = b2luigi.Parameter()
    file_name = b2luigi.Parameter()

    def output(self):
        yield {"input.root": b2luigi.LocalTarget(_build_data_path(self))}


class RawDataTask(DataTask):
    data_mode = DataMode.raw


class DstDataTask(DataTask):
    release = b2luigi.Parameter()
    prod = b2luigi.IntParameter()
    database = b2luigi.IntParameter()


class MdstDataTask(DstDataTask):
    data_mode = DataMode.mdst


class CdstDataTask(DstDataTask):
    data_mode = DataMode.cdst




requires_raw_data = b2luigi.requires(RawDataTask)
requires_mdst_data = b2luigi.requires(MdstDataTask)
requires_cdst_data = b2luigi.requires(CdstDataTask)



def _build_data_path(parameters):
    mode = parameters.data_mode

    if mode == DataMode.raw:
        return RAW_DIR_STRUCTURE.format(p=parameters)
    elif mode == DataMode.cdst:
        return CDST_DIR_STRUCTURE.format(p=parameters)
    elif mode == DataMode.mdst:
        return MDST_DIR_STRUCTURE.format(p=parameters)


def _parse_data_path(data_mode, data_path):
    extracted_kwargs = None
    if data_mode == DataMode.mdst:
        extracted_kwargs = parse(MDST_DIR_STRUCTURE, data_path)
    elif data_mode == DataMode.cdst:
        extracted_kwargs = parse(CDST_DIR_STRUCTURE, data_path)
    elif data_mode == DataMode.raw:
        extracted_kwargs = parse(RAW_DIR_STRUCTURE, data_path)

    if not extracted_kwargs:
        raise ValueError("The path does not fit the given format")
    extracted_kwargs = {key.replace("p.", ""): value for key, value in extracted_kwargs.named.items()}
    return extracted_kwargs


def _get_data_kwargs(data_mode, experiment_number, run_number, prefix=None, file_name=None, **other_kwargs):
    if file_name is None:
        file_name = "*"

    if prefix is None:
        prefix = "*"

    all_kwargs = fill_kwargs_with_lists(data_mode=data_mode, experiment_number=experiment_number, run_number=run_number, prefix=prefix,
                                        file_name=file_name, **other_kwargs)

    for kwargs in product_dict(**all_kwargs):
        # The build_data_path wants an object, so lets convert the dict to a named tuple
        kwargs = namedtuple('GenericDict', kwargs.keys())(**kwargs)
        for data_file in glob(_build_data_path(kwargs)):
            yield _parse_data_path(data_mode, data_file)


class TaskWithDataSupport:
    def clone_on_mdst(self, task_class, experiment_number, run_number, release, prod, database, prefix=None, file_name=None,
                      **additional_kwargs):
        # TODO: make database not needed
        for kwargs in _get_data_kwargs(data_mode=DataMode.mdst, experiment_number=experiment_number, run_number=run_number, release=release,
                                       prod=prod, database=database, prefix=prefix, file_name=file_name):
            yield self.clone(task_class, **kwargs, **additional_kwargs)

    def clone_on_cdst(self, task_class, experiment_number, run_number, release, prod, database, prefix=None, file_name=None,
                      **additional_kwargs):
        # TODO: make database not needed
        for kwargs in _get_data_kwargs(data_mode=DataMode.cdst, experiment_number=experiment_number, run_number=run_number, release=release,
                                       prod=prod, database=database, prefix=prefix, file_name=file_name):
            yield self.clone(task_class, **kwargs, **additional_kwargs)

    def clone_on_raw(self, task_class, experiment_number, run_number, prefix=None, file_name=None, **additional_kwargs):
        for kwargs in _get_data_kwargs(data_mode=DataMode.raw, experiment_number=experiment_number, run_number=run_number, prefix=prefix,
                                       file_name=file_name):
            yield self.clone(task_class, **kwargs, **additional_kwargs)



import contextlib
import importlib.util
import os

import argparse

from collections import defaultdict


@contextlib.contextmanager
def remember_cwd():
    old_cwd = os.getcwd()
    try:
        yield
    finally:
        os.chdir(old_cwd)


def get_all_outputs(task, only_tasks=False):
    if only_tasks:
        yield task.complete(), task
    else:
        outputs = task.output()
        if isinstance(outputs, dict):
            outputs = outputs.values()
        for output in outputs:
            yield output.exists(), output.path

    requirements = task.requires()

    try:
        for requirement in requirements:
            yield from get_all_outputs(requirement, only_tasks=only_tasks)
    except TypeError:
        yield from get_all_outputs(requirements, only_tasks=only_tasks)
    


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("file_name", help="Which file name to check.")
    parser.add_argument("--task_name", help="Which task name to check", default="MainTask")
    parser.add_argument("--only-tasks", help="Check only tasks", action="store_true")
    parser.add_argument("--only-missing", help="Show only missing entities", action="store_true")

    args = parser.parse_args()

    file_name = args.file_name
    task_class = args.task_name
    only_tasks = args.only_tasks
    only_missing = args.only_missing
    kwargs = {}

    file_name = os.path.abspath(file_name)

    with remember_cwd():
        os.chdir(os.path.dirname(file_name))
        spec = importlib.util.spec_from_file_location("module.name", file_name)
        task_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(task_module)

        m = getattr(task_module, task_class)(**kwargs)

        output_files = defaultdict(set)
        for key, value in get_all_outputs(m, only_tasks=only_tasks):
            output_files[key].add(value)

        for key, file_list in output_files.items():
            if key and only_missing:
                 continue
            try:
                file_list = sorted(file_list)
            except TypeError:
                pass

            if key:
                string = "Existing"
            else:
                string = "Missing"
            if file_list:
                print(f"{string} targets:")
                for output_file in file_list:
                    print(output_file)
            else:
                print(f"No {string.lower()} targets!")

            print()
